#import "PSFPController.h"

#define TableRowType @"row"
#define TableRowTypes [NSArray arrayWithObjects:@"row",nil]

@implementation PSFPController

- (id)init
{
	[super init];
	player = [[PSFPlayer alloc] initWithDelegate: self];
	playListArray = [[NSMutableArray alloc] init];
	playListStatusArray = [[NSMutableArray alloc] init];
	playListTitleArray = [[NSMutableArray alloc] init];
	nowPlaying = NO;
	nowConverting = NO;
	launched = NO;
	return self;
}

- (IBAction)play:(id)sender
{
	if(!nowPlaying) {
		if([playListArray count]) {
			[self playFileAtIndex: 0];
			nowPlaying = YES;
		}
		else return;
	}
	else {
		if([player isPause]) [playButton setImage:[NSImage imageNamed:@"pause"]];
		else [playButton setImage:[NSImage imageNamed:@"play"]];
		[player pause];
	}
}

- (IBAction)stop:(id)sender
{
	if(nowPlaying) {
		[player stop];
		nowPlaying = NO;
		[playingTime setStringValue:@""];
		[statusText setStringValue: @""];
		[currentFile setStringValue: @""];
		if(playingFile) [playingFile release];
		[self clearStatus];
		[playListTable reloadData];
		[convertMenu setEnabled: YES];
		[playButton setImage:[NSImage imageNamed:@"play"]];
		nowPlaying = NO;
	}
	else if(nowConverting) {
		int index = [playListArray indexOfObjectIdenticalTo: [converter currentFile]];
		[playListStatusArray replaceObjectAtIndex: index withObject: [NSNumber numberWithInt: 2]];
		[converter stop];
	}
}

- (IBAction)onDoubleClick:(id)sender
{
	if(nowConverting) return;
	if([sender clickedRow] < 0) return;
	if(nowPlaying) {
		int index = [playListArray indexOfObjectIdenticalTo: playingFile];
		[playListStatusArray replaceObjectAtIndex: index withObject: [NSNumber numberWithInt: 0]];
	}
	[self playFileAtIndex: [sender clickedRow]];
	nowPlaying = YES;
}

- (IBAction)skip:(id)sender
{
	if(!nowPlaying) return;
	int index = [playListArray indexOfObjectIdenticalTo: playingFile];
	[playListStatusArray replaceObjectAtIndex: index withObject: [NSNumber numberWithInt: 0]];
	if(index >= [playListArray count]-1) index = -1;
	[self playFileAtIndex: index+1];
	[playListTable scrollRowToVisible: index+1];
	[playListTable reloadData];
}

- (IBAction)reverse:(id)sender
{
	if(!nowPlaying) return;
	int index = [playListArray indexOfObjectIdenticalTo: playingFile];
	[playListStatusArray replaceObjectAtIndex: index withObject: [NSNumber numberWithInt: 0]];
	if(index == 0) index = [playListArray count];
	[self playFileAtIndex: index-1];
	[playListTable scrollRowToVisible: index-1];
	[playListTable reloadData];
}

- (IBAction)setChannelMuteSPC:(id)sender
{
	[player toggleMuteChannelSPC: [sender tag]];
}

- (IBAction)setChannelMutePSF:(id)sender
{
	[player toggleMuteChannelPSF: [sender tag]];
}

- (void)playFileAtIndex:(int)index
{
	FILE *fp;
	char buf[4];
	if(!(fp=fopen([[playListArray objectAtIndex: index] UTF8String],"rb"))) {
		return;
	}
	if(fread(buf,1,4,fp)!=4) {
		fclose(fp);
		return;
	}
	fclose(fp);
	[convertMenu setEnabled: NO];
	[playButton setImage:[NSImage imageNamed:@"pause"]];
	
	if(!memcmp(buf,"PSF\x01",4)) {
		[player playPSF: [playListArray objectAtIndex: index]];
	}
	else if(!memcmp(buf,"SNES",4)) {
		[player playSPC: [playListArray objectAtIndex: index]];
	}
	
	[self clearStatus];
	if(nowPlaying) [playingFile release];
	playingFile = [[playListArray objectAtIndex: index] retain];
	[statusText setStringValue: @"Playing"];
	[playingTime setStringValue:@""];
	[currentFile setStringValue: [playListTitleArray objectAtIndex: index]];
	[playListStatusArray replaceObjectAtIndex: index withObject: [NSNumber numberWithInt: 3]];
	[playListTable reloadData];
}

- (void)playDone
{
	if(nowPlaying) {
		int index = [playListArray indexOfObjectIdenticalTo: playingFile];
		[playListStatusArray replaceObjectAtIndex: index withObject: [NSNumber numberWithInt: 0]];
		if(index >= [playListArray count]-1) index = -1;
		[self playFileAtIndex: index+1];
		[playListTable scrollRowToVisible: index+1];
		[playListTable reloadData];
	}
}

- (void)clearStatus
{
	int n = [playListArray count];
	int i;
	[playListStatusArray removeAllObjects];
	for(i=0;i<n;i++) {
		[playListStatusArray addObject: [NSNumber numberWithInt: 0]];
	}
}

- (IBAction)convert:(id)sender
{
	int i,n;
	n = [playListArray count];
	NSMutableArray *ar = [NSMutableArray arrayWithCapacity: 0];
	for(i=0;i<n;i++) {
		if([playListTable isRowSelected:i]) {
			[ar addObject: [playListArray objectAtIndex: i]];
		}
	}
	if(![ar count]) return;
	[self clearStatus];
	[playButton setEnabled: NO];
	[skipButton setEnabled: NO];
	[reverseButton setEnabled: NO];
	[convertMenu setEnabled: NO];
	converter = [[PSFPConverter alloc] initWithDelegate: self];
	if([[dupRadio selectedCell] tag] == 0) [converter setAllowDuplicatedName: NO];
	else [converter setAllowDuplicatedName: YES];
	if([[outputRadio selectedCell] tag] == 1) [converter setOutputDirectory: [fixedOutputDir stringValue]];
	[converter setConvertFiles: ar];
	[converter convert];
	[statusText setStringValue: @"Converting"];
	[currentFile setStringValue: [playListTitleArray objectAtIndex: [playListArray indexOfObjectIdenticalTo: [converter currentFile]]]];
	int index = [playListArray indexOfObjectIdenticalTo: [converter currentFile]];
	[playListStatusArray replaceObjectAtIndex: index withObject: [NSNumber numberWithInt: 4]];
	[playListTable reloadData];
	nowConverting = YES;
}

- (void)convertDone:(NSString *)doneFile
{
	NSString *file = [converter currentFile];
	[currentFile setStringValue: [file lastPathComponent]];
	int index = [playListArray indexOfObjectIdenticalTo: doneFile];
	[playListStatusArray replaceObjectAtIndex: index withObject: [NSNumber numberWithInt: 1]];
	index = [playListArray indexOfObjectIdenticalTo: file];
	[playListStatusArray replaceObjectAtIndex: index withObject: [NSNumber numberWithInt: 3]];
	[playListTable reloadData];
}

- (void)convertComplete:(NSString *)doneFile
{
	int index = [playListArray indexOfObjectIdenticalTo: doneFile];
	[playListStatusArray replaceObjectAtIndex: index withObject: [NSNumber numberWithInt: 1]];
	[playingTime setStringValue:@""];
	[statusText setStringValue: @""];
	[currentFile setStringValue: @""];
	[converter release];
	[playListTable reloadData];
	[playButton setEnabled: YES];
	[skipButton setEnabled: YES];
	[reverseButton setEnabled: YES];
	[convertMenu setEnabled: YES];
	nowConverting = NO;
}

- (void)convertTerminated
{
	nowConverting = NO;
	[playingTime setStringValue:@""];
	[statusText setStringValue: @""];
	[currentFile setStringValue: @""];
	[playListTable reloadData];
	[playButton setEnabled: YES];
	[skipButton setEnabled: YES];
	[reverseButton setEnabled: YES];
	[converter release];
	[convertMenu setEnabled: YES];
}

- (void)setPlayTime:(NSString *) time
{
	[playingTime setStringValue:time];
}

- (IBAction)prefCancel:(id)sender
{
	[prefPane close];
	[self loadPrefs];
}

- (IBAction)prefOK:(id)sender
{
	[prefPane close];
	[self savePrefs];
}

- (IBAction)showPref:(id)sender
{
	[prefPane center];
	[prefPane makeKeyAndOrderFront:self];
}

- (IBAction)enableOutputDir:(id)sender
{
	if([[sender selectedCell] tag] == 1) {
		[fixedOutputDir setEnabled: YES];
		[setOutputDirButton setEnabled:YES];
	}
	else {
		[fixedOutputDir setEnabled: NO];
		[setOutputDirButton setEnabled:NO];
	}
}

- (IBAction)setOutputDir:(id)sender
{
	NSOpenPanel *op;
	int stat;
	
	op = [ NSOpenPanel openPanel ];
	
	//[ op setTitle : NSLocalizedString(@"selectdir",@"") ];
	[ op setCanChooseDirectories : YES ];
	[ op setCanChooseFiles : NO ];
	
	stat = [ op runModalForDirectory : nil
								file : nil
							   types : nil ];
							   
	if(stat != NSOKButton) return;
	
	[fixedOutputDir setStringValue:[op filename]];
}

- (IBAction)toggleUnlimited:(id)sender
{
	if([sender state] == NSOnState) {
		[depthStepper setEnabled: NO];
		[subdirectoryDepth setTextColor: [NSColor lightGrayColor]];
	}
	else {
		[depthStepper setEnabled: YES];
		[subdirectoryDepth setTextColor: [NSColor blackColor]];
	}
}

- (IBAction)setWindowOpacity:(id)sender
{
	[mainWindow setAlphaValue: [sender floatValue]];
}

- (IBAction)toggleWindowAppearance:(id)sender
{
	if([sender state] == NSOnState) {
		[mainWindow _setTexturedBackground: YES];
		[channelPane _setTexturedBackground: YES];
	}
	else {
		[mainWindow _setTexturedBackground: NO];
		[channelPane _setTexturedBackground: NO];
	}
}

- (IBAction)setWindowLevel:(id)sender
{
	if([[sender selectedItem] tag] == 0) {
		[mainWindow setLevel: NSNormalWindowLevel];
	}
	else {
		[mainWindow setLevel: NSFloatingWindowLevel];
	}
}

- (IBAction)toggleTitle:(id)sender
{
	[playListTable reloadData];
}

- (IBAction)showChannelPane:(id)sender
{
	[channelPane makeKeyAndOrderFront:self];
}

- (void)savePrefs
{
	NSUserDefaults *pref = [NSUserDefaults standardUserDefaults];
	[pref setObject:[fixedOutputDir stringValue] forKey:@"OutputDir"];
	[pref setInteger:[[dupRadio selectedCell] tag] forKey:@"DupRadio"];
	[pref setInteger:[[outputRadio selectedCell] tag] forKey:@"OutputRadio"];
	if([autoSaveFlag state] == NSOnState) [pref setObject:playListArray forKey:@"PlayList"];
	else [pref removeObjectForKey: @"PlayList"];
	[pref setInteger: [autoSaveFlag state] forKey: @"AutoSave"];
	[pref setInteger: [subdirectoryDepth intValue] forKey : @"Depth"];
	[pref setInteger: [unlimitedCheck state] forKey: @"Unlimited"];
	[pref setFloat: [opacitySlider floatValue] forKey: @"Opacity"];
	[pref setInteger: [texturedCheck state] forKey: @"Texture"];
	[pref setInteger:[[windowLevelPopup selectedItem] tag] forKey:@"WindowLevel"];
	[pref setInteger: [titleCheckBox state] forKey: @"ShowTitle"];
	[pref setObject:[playTimeField stringValue] forKey:@"PlayTime"];
	[pref setObject:[fadeOutField stringValue] forKey:@"FadeOut"];
	[pref setInteger:[[playTimeRadio selectedCell] tag] forKey:@"PlayTimeRadio"];
	[pref setObject: [channelPane stringWithSavedFrame] forKey:@"ChannelPaneFrame"];
	[pref setInteger: [reverbEnabledSPC state] forKey: @"ReverbEnabledSPC"];
	[pref setInteger: [reverbEnabledPSF state] forKey: @"ReverbEnabledPSF"];
	[pref setInteger:[[reverbType selectedItem] tag] forKey:@"ReverbType"];
	[pref synchronize];
}

- (void)loadPrefs
{
	NSUserDefaults *pref = [NSUserDefaults standardUserDefaults];
	id temp;
	int i;
	if(temp=[pref objectForKey:@"DupRadio"]) {
		i = [temp intValue];
		[dupRadio selectCellWithTag: i];
	}
	if(temp=[pref objectForKey:@"OutputRadio"]) {
		i = [temp intValue];
		[outputRadio selectCellWithTag: i];
		if(i==1) {
			[fixedOutputDir setEnabled: YES];
			[setOutputDirButton setEnabled:YES];
		}
		else {
			[fixedOutputDir setEnabled: NO];
			[setOutputDirButton setEnabled:NO];
		}
	}
	if(temp=[pref objectForKey:@"OutputDir"]) {
		[fixedOutputDir setStringValue:temp];
	}
	if(temp=[pref objectForKey:@"AutoSave"]) {
		[autoSaveFlag setState: [temp intValue]];
	}
	else [fixedOutputDir setStringValue:[@"~/Desktop" stringByExpandingTildeInPath]];
	if(temp=[pref objectForKey:@"Depth"]) {
		[subdirectoryDepth setIntValue: [temp intValue]];
		[depthStepper setIntValue: [temp intValue]];
	}
	if(temp=[pref objectForKey:@"Unlimited"]) {
		[unlimitedCheck setState: [temp intValue]];
		if([unlimitedCheck state] == NSOnState) {
			[depthStepper setEnabled: NO];
			[subdirectoryDepth setTextColor: [NSColor lightGrayColor]];
		}
		else {
			[depthStepper setEnabled: YES];
			[subdirectoryDepth setTextColor: [NSColor blackColor]];
		}
	}
	if(temp=[pref objectForKey:@"Opacity"]) {
		[mainWindow setAlphaValue: [temp floatValue]];
		[opacitySlider setFloatValue: [temp floatValue]];
	}
	if(temp=[pref objectForKey:@"Texture"]) {
		[texturedCheck setState: [temp intValue]];
		if([texturedCheck state] == NSOnState) {
			[mainWindow _setTexturedBackground: YES];
			[channelPane _setTexturedBackground: YES];
		}
		else {
			[mainWindow _setTexturedBackground: NO];
			[channelPane _setTexturedBackground: NO];
		}
	}
	if(temp=[pref objectForKey:@"WindowLevel"]) {
		[windowLevelPopup selectItemAtIndex: [windowLevelPopup indexOfItemWithTag:[temp intValue]]];
		if([temp intValue] == 0) {
			[mainWindow setLevel: NSNormalWindowLevel];
		}
		else {
			[mainWindow setLevel: NSFloatingWindowLevel];
		}
	}
	if(temp=[pref objectForKey:@"PlayTime"]) {
		[playTimeField setStringValue:temp];
	}
	if(temp=[pref objectForKey:@"FadeOut"]) {
		[fadeOutField setStringValue:temp];
	}
	if(temp=[pref objectForKey:@"PlayTimeRadio"]) {
		i = [temp intValue];
		[playTimeRadio selectCellWithTag: i];
	}
	if(temp=[pref objectForKey:@"ReverbEnabledSPC"]) {
		[reverbEnabledSPC setState: [temp intValue]];
	}
	if(temp=[pref objectForKey:@"ReverbEnabledPSF"]) {
		[reverbEnabledPSF setState: [temp intValue]];
	}
	if(temp=[pref objectForKey:@"ReverbType"]) {
		i = [temp intValue];
		[reverbType selectItemAtIndex: i];
	}
	if(launched) return;
	if(temp=[pref objectForKey:@"PlayList"]) {
		[self appendFile:temp atIndex: 0 recursion: 0];
	}
	if(temp=[pref objectForKey:@"ShowTitle"]) {
		[titleCheckBox setState: [temp intValue]];
	}
	if(temp = [pref objectForKey: @"ChannelPaneFrame"]) {
		[channelPane setFrameFromString: temp];
	}
}

- (void)removeSelectedFile:(id)table
{
	int n,m,i;
	
	if(![playListArray count]) return;
	if(nowConverting) return;
	
	n = [playListArray count];
	if(nowPlaying) m = [playListArray indexOfObjectIdenticalTo: playingFile];
	else m = -1;
	for(i=n-1;i>=0;i--) {
		if([playListTable isRowSelected:i]) {
			if(i == m) continue;
			[playListArray removeObjectAtIndex:i];
			[playListStatusArray removeObjectAtIndex:i];
			[playListTitleArray removeObjectAtIndex:i];
			//if(i < [statusArray count]) [statusArray removeObjectAtIndex:i];
		}
	}
	[playListTable reloadData];
	[playListTable deselectAll:nil];
}

- (int)appendFile:(NSArray *)files atIndex: (int)index recursion: (int)r
{
	int i, j, added=0, ret, m;
	int n = [files count];
	FILE *fp;
	char buf[4];
	NSFileManager *fm = [NSFileManager defaultManager];
	BOOL isDir;
	NSString *file;
	/*
	for(i=0;i<n;i++) {
		NSLog(@"hoge");
		NSRect myRect = [playListTable frameOfCellAtColumn:0 row: m+i];
		NSProgressIndicator *myCheck = [[NSProgressIndicator alloc] initWithFrame: myRect];
		//[myCheck setButtonType: NSSwitchButton];
		[playListConvertArray addObject:[NSNumber numberWithInt: 1]];
		[playListCheckBoxArray addObject: myCheck];
		[playListTable addSubview: myCheck];
		[myCheck release];
	}
	*/
	for(i=0;i<n;i++) {
		file = [files objectAtIndex: i];
		[fm fileExistsAtPath:file isDirectory:&isDir];
		if(isDir) {
			if([unlimitedCheck state] == NSOffState && r >= [subdirectoryDepth intValue]) continue;
			NSArray *arr = [fm directoryContentsAtPath: file];
			NSMutableArray *arr2 = [NSMutableArray arrayWithArray: 0];
			m = [arr count];
			for(j=0;j<m;j++) {
				[arr2 addObject: [file stringByAppendingPathComponent: [arr objectAtIndex: j]]];
			}
			ret = [self appendFile: arr2 atIndex: index+added recursion: r+1];
			added = added + ret;
			continue;
		}
		
		if([[file pathExtension] caseInsensitiveCompare: @"psflib"] == NSOrderedSame) continue;
		
		if(!(fp=fopen([file UTF8String],"rb"))) {
			continue;
		}
		if(fread(buf,1,4,fp)!=4) {
			fclose(fp);
			continue;
		}
		fclose(fp);
		if(memcmp(buf,"PSF\x01",4) && memcmp(buf,"SNES",4))	// Only allow for PSF1 files for now.
		{
			continue;
		}
		[playListArray insertObject: file atIndex: index+added];
		[playListStatusArray insertObject: [NSNumber numberWithInt: 0] atIndex: index+added];
		[playListTitleArray insertObject: [self getMusicTitle:file] atIndex: index+added];
		added++;
	}
	[playListTable reloadData];
	return added;
}

- (NSString *)getMusicTitle:(NSString *)str
{
	FILE *fp;
	char buf[64];
	char title[128];
	char game[128];
	int i,j=0;
	title[0] = '\0';
	game[0] = '\0';
	fp = fopen([str UTF8String], "rb");
	fread(buf, 1, 4, fp);
	if(!memcmp(buf,"SNES",4)) {
		fseek(fp,0x23,SEEK_SET);
		char c;
		fread(&c,1,1,fp);
		if(c == 0x1a)  {
			fseek(fp, 0x2e, SEEK_SET);
			fread(title, 1, 32, fp);
			title[32] = '\0';
			fread(game, 1, 32, fp);
			game[32] = '\0';
		}
		if(fseek(fp, 0x10200, SEEK_SET) != 0) j=1;
		if(fread(buf, 1, 4, fp) != 4) j=1;
		if(!memcmp(buf,"xid6",4) && !j) {
			fseek(fp, 4, SEEK_CUR);
			while(1) {
				char id;
				char type;
				short length;
				if(fread(&id, 1, 1, fp) != 1) break;
				if(fread(&type, 1, 1, fp) != 1) break;
				if(id == 1) {
					fread(&length, 2, 1, fp);
					length = ((length >> 8) & 0xFF) | ((length << 8) & 0xFF00);
					fread(title, 1, length, fp);
				}
				else if(id == 2) {
					fread(&length, 2, 1, fp);
					length = ((length >> 8) & 0xFF) | ((length << 8) & 0xFF00);
					fread(game, 1, length, fp);
				}
				else if(id == 0) {
					fseek(fp, -1, SEEK_CUR);
				}
				else {
					switch(type) {
					  case 0:
						fseek(fp, 2, SEEK_CUR);
						break;
					  case 1:
						fread(&length, 2, 1, fp);
						length = ((length >> 8) & 0xFF) | ((length << 8) & 0xFF00);
						fseek(fp, length, SEEK_CUR);
						break;
					  case 4:
						fseek(fp, 6, SEEK_CUR);
						break;
					  default:
						break;
					}
				}
			}
		}
	}
	else if(!memcmp(buf,"PSF\x01",4)) {
		int size = 0, tmp;
		fseek(fp, 4, SEEK_SET);
		fread(&tmp, 4, 1, fp);
		size += ((tmp >> 24) & 0xFF) | ((tmp >> 8) & 0xFF00) | ((tmp << 8) & 0xFF0000) | ((tmp << 24) & 0xFF000000);
		fread(&tmp, 4, 1, fp);
		size += ((tmp >> 24) & 0xFF) | ((tmp >> 8) & 0xFF00) | ((tmp << 8) & 0xFF0000) | ((tmp << 24) & 0xFF000000);
		fseek(fp, 4+size, SEEK_CUR);
		if(fread(buf, 1, 5, fp) != 5) j=1;
		if(memcmp(buf, "[TAG]", 5)) j=1;
		i=0;
		if(!j) {
		while(1) {
			char c;
			if(fread(&c, 1, 1, fp) != 1) break;
			if(c=='=') {
				fseek(fp, 0-i-1, SEEK_CUR);
				fread(buf, 1, i, fp);
				buf[i] = '\0';
				fseek(fp,1, SEEK_CUR);
				if(!strcasecmp(buf, "title")) {
					i=0;
					while(1) {
						fread(&c, 1, 1, fp);
						title[i] = c;
						if(c==0xa) break;
						i++;
					}
					title[i] = '\0';
				}
				else if(!strcasecmp(buf, "game")) {
					i=0;
					while(1) {
						fread(&c, 1, 1, fp);
						game[i] = c;
						if(c==0xa) break;
						i++;
					}
					game[i] = '\0';
				}
				else {
					while(1) {
						fread(&c, 1, 1, fp);
						if(c==0xa) break;
					}
				}
				i=0;
				continue;
			}
			i++;
		}
	}
	}
	fclose(fp);
	
	if(title[0]) {
		if(game[0]) return [NSString stringWithFormat: @"%s - %s", game, title];
		else return [NSString stringWithFormat: @"%s", title];
	}
	else return [str lastPathComponent];
}

- (NSString *)playTime
{
	return [playTimeField stringValue];
}

- (NSString *)fadeOut
{
	return [fadeOutField stringValue];
}

- (BOOL)forcePlayTime
{
	if([[playTimeRadio selectedCell] tag] == 0) return NO;
	else return YES;
}

- (int)reverbType
{
	[[reverbType selectedItem] tag];
}

- (BOOL)reverbEnabledSPC
{
	if([reverbEnabledSPC state] == NSOnState) return YES;
	else return NO;
}

- (BOOL)reverbEnabledPSF
{
	if([reverbEnabledPSF state] == NSOnState) return YES;
	else return NO;
}

/* delegate methods */

- (int)numberOfRowsInTableView:(NSTableView *)tableView
{
	return [playListArray count];
}

- (id)tableView:(NSTableView *)tableView
		objectValueForTableColumn:(NSTableColumn *)tableColumn
		row:(int)row
{
	id identifier = [tableColumn identifier];
	if([identifier isEqual:@"Status"])  {
		if([[playListStatusArray objectAtIndex:row] intValue] == 1) return [NSImage imageNamed:@"done"];
		else if([[playListStatusArray objectAtIndex:row] intValue] == 2) return [NSImage imageNamed:@"error"];
		else if([[playListStatusArray objectAtIndex:row] intValue] == 3) return [NSImage imageNamed:@"speaker"];
		else if([[playListStatusArray objectAtIndex:row] intValue] == 4) return [NSImage imageNamed:@"process"];
		else return nil;
	}
	else if([identifier isEqual:@"File"]){
		if([titleCheckBox state] == NSOnState) return [[playListArray objectAtIndex:row] lastPathComponent];
		else return [playListTitleArray objectAtIndex:row];
	}
	return nil;
}


-(BOOL)tableView:(NSTableView *)tv writeRows:(NSArray*)rows toPasteboard:(NSPasteboard*)pboard
{
	
	[pboard declareTypes:TableRowTypes owner:self];
	[pboard setPropertyList:rows forType:TableRowType];
	
	return YES;
}

-(NSDragOperation)tableView:(NSTableView*)tv validateDrop:(id <NSDraggingInfo>)info proposedRow:(int)row proposedDropOperation:(NSTableViewDropOperation)op
{
	if (op == NSTableViewDropAbove ) {
		return NSDragOperationGeneric;
	} else {
		return NSDragOperationNone;
	}
}

-(BOOL)tableView:(NSTableView*)tv acceptDrop:(id <NSDraggingInfo>)info row:(int)row dropOperation:(NSTableViewDropOperation)op
{

	NSPasteboard *pboard = [info draggingPasteboard];
	NSArray *ar = [pboard propertyListForType: NSFilenamesPboardType];
	if(ar) {
		NSArray *sortedArray = [ar sortedArrayUsingSelector:@selector(caseInsensitiveCompare:)];
		[self appendFile:sortedArray atIndex: row recursion: 0];
		return YES;
	}
	
	NSEnumerator *e=[[pboard propertyListForType:TableRowType] objectEnumerator];
	NSNumber *number;
	NSMutableArray *upperArray1=[NSMutableArray arrayWithArray:[playListArray subarrayWithRange:NSMakeRange(0,row)]];
	NSMutableArray *lowerArray1=[NSMutableArray arrayWithArray:[playListArray subarrayWithRange:NSMakeRange(row,([playListArray count] - row))]];
	NSMutableArray *middleArray1=[NSMutableArray arrayWithCapacity:0];
	NSMutableArray *upperArray2=[NSMutableArray arrayWithArray:[playListStatusArray subarrayWithRange:NSMakeRange(0,row)]];
	NSMutableArray *lowerArray2=[NSMutableArray arrayWithArray:[playListStatusArray subarrayWithRange:NSMakeRange(row,([playListArray count] - row))]];
	NSMutableArray *middleArray2=[NSMutableArray arrayWithCapacity:0];
	NSMutableArray *upperArray3=[NSMutableArray arrayWithArray:[playListTitleArray subarrayWithRange:NSMakeRange(0,row)]];
	NSMutableArray *lowerArray3=[NSMutableArray arrayWithArray:[playListTitleArray subarrayWithRange:NSMakeRange(row,([playListArray count] - row))]];
	NSMutableArray *middleArray3=[NSMutableArray arrayWithCapacity:0];
	id object1,object2,object3;
	int i;

	if (op == NSTableViewDropAbove && [pboard availableTypeFromArray:TableRowTypes] != nil) {

		while ((number=[e nextObject]) != nil) {
			object1=[playListArray objectAtIndex:[number intValue]];
			object2=[playListStatusArray objectAtIndex:[number intValue]];
			object3=[playListTitleArray objectAtIndex:[number intValue]];
			[middleArray1 addObject:object1];
			[middleArray2 addObject:object2];
			[middleArray3 addObject:object3];
			i = [upperArray1 indexOfObjectIdenticalTo: object1];
			if(i != NSNotFound) {
				[upperArray1 removeObjectAtIndex: i];
				[upperArray2 removeObjectAtIndex: i];
				[upperArray3 removeObjectAtIndex: i];
			}
			i = [lowerArray1 indexOfObjectIdenticalTo: object1];
			if(i != NSNotFound) {
				[lowerArray1 removeObjectAtIndex: i];
				[lowerArray2 removeObjectAtIndex: i];
				[lowerArray3 removeObjectAtIndex: i];
			}
		}

		[playListArray removeAllObjects];
		[playListStatusArray removeAllObjects];
		[playListTitleArray removeAllObjects];

		[playListArray addObjectsFromArray:upperArray1];
		[playListArray addObjectsFromArray:middleArray1];
		[playListArray addObjectsFromArray:lowerArray1];
		[playListStatusArray addObjectsFromArray:upperArray2];
		[playListStatusArray addObjectsFromArray:middleArray2];
		[playListStatusArray addObjectsFromArray:lowerArray2];
		[playListTitleArray addObjectsFromArray:upperArray3];
		[playListTitleArray addObjectsFromArray:middleArray3];
		[playListTitleArray addObjectsFromArray:lowerArray3];

		[playListTable reloadData];
		[playListTable deselectAll:nil];
	
		for (i=[upperArray1 count];i<([upperArray1 count] + [middleArray1 count]);i++) {
			[playListTable selectRow:i byExtendingSelection:[playListTable allowsMultipleSelection]];
		}

		return YES;
	} else {
		return NO;
	}
	
	return YES;
}


- (void)applicationDidFinishLaunching: (NSNotification *)notification
{
	[self loadPrefs];
	NSArray* array = [NSArray arrayWithObjects:NSFilenamesPboardType, @"row", nil];
	[playListTable registerForDraggedTypes:array];
	[[playListTable tableColumnWithIdentifier:@"Status"] setDataCell:[[NSImageCell alloc]init]];
	[playListTable setDoubleAction: @selector(onDoubleClick:)];
	[mainWindow setFrameAutosaveName: @"MainWindow"];
	[mainWindow makeKeyAndOrderFront: self];
	BOOL canWeChangeTextureOnTheFly = [mainWindow respondsToSelector:@selector(_setTexturedBackground:)];
	if(!canWeChangeTextureOnTheFly) [texturedCheck setEnabled: NO];
	launched = YES;
}


- (NSApplicationTerminateReply)applicationShouldTerminate:(NSApplication *)sender
{
	if(nowConverting || nowPlaying) {
		int ret = NSRunAlertPanel(@"Quit", @"Some files are being processed. Quit anyway?",@"OK", nil, @"Cancel");
		if(ret == NSAlertOtherReturn) return NO;
	}
	[self savePrefs];
	return YES;
}

- (BOOL)windowShouldClose:(id)sender
{
	[NSApp terminate:nil];
	return NO;
}

- (NSString *)createChannelMaskforSPC
{
	unsigned char mask = 0;
	if([spcCh1 state] == NSOffState) mask |= 1;
	if([spcCh2 state] == NSOffState) mask |= 2;
	if([spcCh3 state] == NSOffState) mask |= 4;
	if([spcCh4 state] == NSOffState) mask |= 8;
	if([spcCh5 state] == NSOffState) mask |= 16;
	if([spcCh6 state] == NSOffState) mask |= 32;
	if([spcCh7 state] == NSOffState) mask |= 64;
	if([spcCh8 state] == NSOffState) mask |= 128;
	return [NSString stringWithFormat: @"%u", mask];
}

- (NSString *)createChannelMaskforPSF
{
	int mask = 0;
	if([psfCh1 state] == NSOffState) mask |= 1;
	if([psfCh2 state] == NSOffState) mask |= 2;
	if([psfCh3 state] == NSOffState) mask |= 4;
	if([psfCh4 state] == NSOffState) mask |= 8;
	if([psfCh5 state] == NSOffState) mask |= 16;
	if([psfCh6 state] == NSOffState) mask |= 32;
	if([psfCh7 state] == NSOffState) mask |= 64;
	if([psfCh8 state] == NSOffState) mask |= 128;
	if([psfCh9 state] == NSOffState) mask |= 256;
	if([psfCh10 state] == NSOffState) mask |= 512;
	if([psfCh11 state] == NSOffState) mask |= 1024;
	if([psfCh12 state] == NSOffState) mask |= 2048;
	if([psfCh13 state] == NSOffState) mask |= 4096;
	if([psfCh14 state] == NSOffState) mask |= 8192;
	if([psfCh15 state] == NSOffState) mask |= 16384;
	if([psfCh16 state] == NSOffState) mask |= 32768;
	if([psfCh17 state] == NSOffState) mask |= 65536;
	if([psfCh18 state] == NSOffState) mask |= 131072;
	if([psfCh19 state] == NSOffState) mask |= 262144;
	if([psfCh20 state] == NSOffState) mask |= 524288;
	if([psfCh21 state] == NSOffState) mask |= 1048576;
	if([psfCh22 state] == NSOffState) mask |= 2097152;
	if([psfCh23 state] == NSOffState) mask |= 4194304;
	if([psfCh24 state] == NSOffState) mask |= 8388608;
	return [NSString stringWithFormat: @"%d", mask];
}


@end
